"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.bulkSendRawTransactions = exports.bulkSendTransactions = exports.bufferToTransaction = exports.stringToTransaction = exports.sendAndConfirmWithRetry = exports.awaitTransactionSignatureConfirmation = exports.sendMultipleInstructionsWithPrioFee = exports.sendAndConfirmWithRetryAndPrioFees = exports.getPriorityFeeEstimate = exports.executeBig = exports.execute = exports.sendMultipleInstructions = exports.sendInstructions = exports.Tx = void 0;
const web3_js_1 = require("@solana/web3.js");
const accountFetchCache_1 = require("./accountFetchCache");
const bs58_1 = __importDefault(require("bs58"));
const anchorError_1 = require("./anchorError");
const logger_1 = require("./logger");
class Tx {
    constructor(p) {
        this.instrs = [];
        this._startPrio = 'NONE';
        this._maxPrio = 'VeryHigh';
        this._commitment = 'confirmed';
        this._addressLookupTableAccounts = [];
        this._sendOptions = { skipPreflight: true };
        this._timeoutMillis = HIGH_END_TIMEOUT_MILLIS;
        this._log = logger_1.consoleLogger;
        this._doNotRetryAt = new Map();
        this._prioFeeEstimateOpts = {
            includeVote: false,
            recommended: true,
        };
        this._signers = [];
        if (p.ixs) {
            this.instrs = p.ixs;
        }
        this._feePayer = p.feePayer;
        if (p.log) {
            this._log = p.log;
        }
    }
    add(...ixs) {
        this.instrs.push(...ixs);
        return this;
    }
    startPrio(p) {
        this._prioFeeEstimateOpts.recommended = false;
        this._startPrio = p;
        return this;
    }
    maxPrio(p) {
        this._maxPrio = p;
        return this;
    }
    commitment(c) {
        this._commitment = c;
        return this;
    }
    lookupTable(as) {
        this._addressLookupTableAccounts = as;
        return this;
    }
    sendOptions(o) {
        this._sendOptions = o;
        return this;
    }
    timeoutMillis(m) {
        this._timeoutMillis = m;
        return this;
    }
    log(l) {
        this._log = l;
        return this;
    }
    doNotRetryAt(errCodeToDesc) {
        this._doNotRetryAt = errCodeToDesc;
        return this;
    }
    prioFeeEstOpts(opts) {
        let opts1 = opts;
        if (this._startPrio != 'NONE') {
            let { recommended } = opts, rest = __rest(opts, ["recommended"]);
            opts1 = rest;
        }
        this._prioFeeEstimateOpts = opts1;
        return this;
    }
    signers(...signers) {
        this._signers.push(...signers);
        return this;
    }
    send(connection) {
        return __awaiter(this, void 0, void 0, function* () {
            const consTx = (instrs) => __awaiter(this, void 0, void 0, function* () {
                const signers = this._signers;
                //if (!(signers.map(s => s.publicKey.toString()).includes(this._feePayer.toString())))
                const blockhash = (yield connection.getLatestBlockhash('confirmed'))
                    .blockhash;
                // feePayer ?? signers[0].publicKey
                const payerKey = this._feePayer;
                if (this._addressLookupTableAccounts.length > 0) {
                    const messageV0 = new web3_js_1.TransactionMessage({
                        payerKey,
                        recentBlockhash: blockhash,
                        instructions: [...instrs],
                    }).compileToV0Message(this._addressLookupTableAccounts);
                    const tx = new web3_js_1.VersionedTransaction(messageV0);
                    tx.sign(signers);
                    return Buffer.from(tx.serialize());
                }
                const tx = new web3_js_1.Transaction({
                    feePayer: payerKey,
                    recentBlockhash: blockhash,
                }).add(...instrs);
                tx.sign(...signers);
                return tx.serialize();
            });
            const recur = (priorityLevel) => __awaiter(this, void 0, void 0, function* () {
                let instrs = this.instrs;
                let txSer = yield consTx(instrs);
                if (priorityLevel !== 'NONE') {
                    const { priorityFeeEstimate } = yield (0, exports.getPriorityFeeEstimate)(connection, txSer, priorityLevel, this._log, this._prioFeeEstimateOpts);
                    if (priorityFeeEstimate) {
                        const computePriceIx = web3_js_1.ComputeBudgetProgram.setComputeUnitPrice({
                            // @ts-ignore
                            microLamports: Math.trunc(priorityFeeEstimate),
                        });
                        txSer = yield consTx([...instrs, computePriceIx]);
                    }
                }
                const log = this._log;
                try {
                    return yield sendAndConfirmWithRetry(connection, txSer, this._sendOptions, this._commitment, this._timeoutMillis, log);
                }
                catch (e) {
                    // @ts-ignore
                    let baseMsg = `failed at prio ${priorityLevel}: ${e} ${e.stack}`;
                    if (priorityLevel == this._maxPrio) {
                        log.info(`${baseMsg} hard failed`);
                        throw e;
                    }
                    const wrappedE = anchorError_1.ProgramError.parse(e, this._doNotRetryAt);
                    if (wrappedE) {
                        log.info(`got do not retry err: ${wrappedE}. aborting`);
                        throw e;
                    }
                    let newPrio;
                    if (priorityLevel == 'Low') {
                        newPrio = 'Medium';
                    }
                    else if (priorityLevel == 'Medium') {
                        newPrio = 'High';
                    }
                    else if (priorityLevel == 'High') {
                        newPrio = 'VeryHigh';
                    }
                    else {
                        newPrio = 'Low';
                    }
                    log.info(`${baseMsg}. retrying at ${newPrio}`);
                    return yield recur(newPrio);
                }
            });
            return yield recur(this._startPrio);
        });
    }
}
exports.Tx = Tx;
function sleep(ms) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve) => setTimeout(resolve, ms));
    });
}
function promiseAllInOrder(it) {
    return __awaiter(this, void 0, void 0, function* () {
        let ret = [];
        for (const i of it) {
            ret.push(yield i());
        }
        return ret;
    });
}
function sendInstructions(provider, instructions, signers = [], payer = provider.wallet.publicKey, commitment = 'confirmed', idlErrors = new Map(), log = logger_1.consoleLogger) {
    return __awaiter(this, void 0, void 0, function* () {
        if (instructions.length == 0) {
            return '';
        }
        let tx = new web3_js_1.Transaction();
        tx.recentBlockhash = (yield provider.connection.getLatestBlockhash()).blockhash;
        tx.feePayer = payer || provider.wallet.publicKey;
        tx.add(...instructions);
        if (signers.length > 0) {
            tx.partialSign(...signers);
        }
        if (tx.feePayer.equals(provider.wallet.publicKey) ||
            tx.instructions.some((ix) => ix.keys.some((key) => key.isSigner && key.pubkey.equals(provider.wallet.publicKey)))) {
            tx = yield provider.wallet.signTransaction(tx);
        }
        try {
            const { txid } = yield sendAndConfirmWithRetry(provider.connection, tx.serialize(), {
                skipPreflight: true,
            }, commitment);
            return txid;
        }
        catch (e) {
            log.error(e);
            if (e.logs) {
                log.error(`Logs: ${e.logs}`);
            }
            const wrappedE = anchorError_1.ProgramError.parse(e, idlErrors);
            throw wrappedE == null ? e : wrappedE;
        }
    });
}
exports.sendInstructions = sendInstructions;
function truthy(value) {
    return !!value;
}
function sendMultipleInstructions(provider, instructionGroups, signerGroups, payer, finality = 'confirmed', idlErrors = new Map(), log = logger_1.consoleLogger) {
    return __awaiter(this, void 0, void 0, function* () {
        const recentBlockhash = (yield provider.connection.getLatestBlockhash('confirmed')).blockhash;
        const ixAndSigners = instructionGroups
            .map((instructions, i) => {
            const signers = signerGroups[i];
            return {
                instructions,
                signers,
            };
        })
            .filter(({ instructions }) => instructions.length > 0);
        const txns = ixAndSigners.map(({ instructions }) => {
            const tx = new web3_js_1.Transaction({
                feePayer: payer || provider.wallet.publicKey,
                recentBlockhash,
            });
            tx.add(...instructions);
            return tx;
        });
        const txnsSignedByWallet = yield provider.wallet.signAllTransactions(txns);
        const txnsSigned = txnsSignedByWallet
            .map((tx, index) => {
            const signers = ixAndSigners[index].signers;
            if (signers.length > 0) {
                tx.partialSign(...signers);
            }
            return tx;
        })
            .map((tx) => tx.serialize());
        log.info('Sending multiple transactions...');
        try {
            return yield promiseAllInOrder(txnsSigned.map((txn) => () => __awaiter(this, void 0, void 0, function* () {
                const { txid } = yield sendAndConfirmWithRetry(provider.connection, txn, {
                    skipPreflight: true,
                }, finality);
                return txid;
            })));
        }
        catch (e) {
            log.error(e);
            const wrappedE = anchorError_1.ProgramError.parse(e, idlErrors);
            throw wrappedE == null ? e : wrappedE;
        }
    });
}
exports.sendMultipleInstructions = sendMultipleInstructions;
function execute(program, provider, command, payer = provider.wallet.publicKey, commitment) {
    var _a;
    return __awaiter(this, void 0, void 0, function* () {
        const { instructions, signers, output } = command;
        const errors = (_a = program.idl.errors) === null || _a === void 0 ? void 0 : _a.reduce((acc, err) => {
            acc.set(err.code, `${err.name}: ${err.msg}`);
            return acc;
        }, new Map());
        if (instructions.length > 0) {
            const txid = yield sendInstructions(provider, instructions, signers, payer, commitment, errors);
            return Object.assign({ txid }, output);
        }
        // @ts-ignore
        return output;
    });
}
exports.execute = execute;
function executeBig(program, provider, command, payer = provider.wallet.publicKey, finality) {
    var _a;
    return __awaiter(this, void 0, void 0, function* () {
        const { instructions, signers, output } = command;
        const errors = (_a = program.idl.errors) === null || _a === void 0 ? void 0 : _a.reduce((acc, err) => {
            acc.set(err.code, `${err.name}: ${err.msg}`);
            return acc;
        }, new Map());
        if (instructions.length > 0) {
            const txids = yield sendMultipleInstructions(provider, instructions, signers, payer || provider.wallet.publicKey, finality, errors);
            return Object.assign(Object.assign({}, output), { txids: Array.from(txids) });
        }
        // @ts-ignore
        return output;
    });
}
exports.executeBig = executeBig;
function getUnixTimeSec() {
    return new Date().valueOf() / 1000;
}
const getPriorityFeeEstimate = (connection, transaction, priorityLevel = 'NONE', log = logger_1.consoleLogger, options = { includeVote: true, recommended: true }) => __awaiter(void 0, void 0, void 0, function* () {
    let opts = options;
    if (!options.recommended) {
        opts = {
            includeVote: options.includeVote,
            priorityLevel: priorityLevel,
        };
    }
    const response = yield fetch(connection.rpcEndpoint, {
        method: 'POST',
        headers: { 'Content-Type': 'application/json' },
        body: JSON.stringify({
            jsonrpc: '2.0',
            id: '1',
            method: 'getPriorityFeeEstimate',
            params: [
                {
                    transaction: bs58_1.default.encode(transaction),
                    options: opts,
                },
            ],
        }),
    });
    if (response.status != 200) {
        const err = `getPriorityFeeEstimate failed ${yield response.text()}`;
        log.error(err);
        throw new Error(err);
    }
    const data = yield response.json();
    let priorityFeeEstimate = 50000;
    if (!data.result || data.result.priorityFeeEstimate == undefined) {
        log.warn(`getPriorityFeeEstimate unknown payload ${JSON.stringify(data)}`);
    }
    else {
        priorityFeeEstimate = data.result.priorityFeeEstimate;
    }
    // @ts-ignore
    log.info(`Fee in function for ${priorityLevel}: ${priorityFeeEstimate}`);
    // @ts-ignore
    return { priorityFeeEstimate };
});
exports.getPriorityFeeEstimate = getPriorityFeeEstimate;
const sendAndConfirmWithRetryAndPrioFees = (connection, instrs, signers, sendOptions, commitment, feePayer, priorityLevel = 'NONE', timeoutMillis = HIGH_END_TIMEOUT_MILLIS, log = logger_1.consoleLogger, addressLookupTableAccounts = [], doNotRetryAt = new Map(), maxPriorityLevel = 'VeryHigh') => __awaiter(void 0, void 0, void 0, function* () {
    const consTx = (instrs) => __awaiter(void 0, void 0, void 0, function* () {
        const blockhash = (yield connection.getLatestBlockhash('confirmed'))
            .blockhash;
        const payerKey = feePayer !== null && feePayer !== void 0 ? feePayer : signers[0].publicKey;
        if (addressLookupTableAccounts.length > 0) {
            const messageV0 = new web3_js_1.TransactionMessage({
                payerKey,
                recentBlockhash: blockhash,
                instructions: [...instrs],
            }).compileToV0Message(addressLookupTableAccounts);
            const tx = new web3_js_1.VersionedTransaction(messageV0);
            tx.sign(signers);
            return Buffer.from(tx.serialize());
        }
        const tx = new web3_js_1.Transaction({
            feePayer: payerKey,
            recentBlockhash: blockhash,
        }).add(...instrs);
        tx.sign(...signers);
        return tx.serialize();
    });
    let txSer = yield consTx(instrs);
    if (priorityLevel !== 'NONE') {
        const feeEstimate = yield (0, exports.getPriorityFeeEstimate)(connection, txSer, priorityLevel, log);
        const computePriceIx = web3_js_1.ComputeBudgetProgram.setComputeUnitPrice({
            // @ts-ignore
            microLamports: Math.trunc(feeEstimate.priorityFeeEstimate),
        });
        txSer = yield consTx([...instrs, computePriceIx]);
    }
    try {
        return yield sendAndConfirmWithRetry(connection, txSer, sendOptions, commitment, timeoutMillis, log);
    }
    catch (e) {
        // @ts-ignore
        let baseMsg = `failed at prio ${priorityLevel}: ${e} ${e.stack}`;
        if (priorityLevel == maxPriorityLevel) {
            log.info(`${baseMsg} hard failed`);
            throw e;
        }
        const wrappedE = anchorError_1.ProgramError.parse(e, doNotRetryAt);
        if (wrappedE) {
            log.info(`got do not retry err: ${wrappedE}. aborting`);
            throw e;
        }
        let newPrio;
        if (priorityLevel == 'Low') {
            newPrio = 'Medium';
        }
        else if (priorityLevel == 'Medium') {
            newPrio = 'High';
        }
        else if (priorityLevel == 'High') {
            newPrio = 'VeryHigh';
        }
        else {
            newPrio = 'Low';
        }
        log.info(`${baseMsg}. retrying at ${newPrio}`);
        return yield (0, exports.sendAndConfirmWithRetryAndPrioFees)(connection, instrs, signers, sendOptions, commitment, feePayer, newPrio, timeoutMillis, log, addressLookupTableAccounts);
    }
});
exports.sendAndConfirmWithRetryAndPrioFees = sendAndConfirmWithRetryAndPrioFees;
const sendMultipleInstructionsWithPrioFee = (provider, instructionGroups, signerGroups, payer, priorityLevel = 'NONE', finality = 'confirmed', timeout = HIGH_END_TIMEOUT_MILLIS, idlErrors = new Map(), log = logger_1.consoleLogger) => __awaiter(void 0, void 0, void 0, function* () {
    const txSigs = instructionGroups.map((_) => '');
    let currI = 0;
    const origPriorityLevel = priorityLevel;
    const sendMultipleTxs = (instructionGroups, signerGroups, priorityLevel) => __awaiter(void 0, void 0, void 0, function* () {
        const recentBlockhash = (yield provider.connection.getLatestBlockhash('confirmed')).blockhash;
        const ixAndSigners = instructionGroups
            .map((instructions, i) => {
            const signers = signerGroups[i];
            return {
                instructions,
                signers,
            };
        })
            .filter(({ instructions }) => instructions.length > 0);
        const txns = ixAndSigners.map(({ instructions }) => {
            const tx = new web3_js_1.Transaction({
                feePayer: payer || provider.wallet.publicKey,
                recentBlockhash,
            });
            tx.add(...instructions);
            return tx;
        });
        const txnsSignedByWallet = yield provider.wallet.signAllTransactions(txns);
        let txnsSigned = txnsSignedByWallet.map((tx, index) => {
            const signers = ixAndSigners[index].signers;
            if (signers.length > 0) {
                tx.partialSign(...signers);
            }
            return tx;
        });
        if (priorityLevel !== 'NONE') {
            const feeEstimates = yield Promise.all(txnsSigned.map((tx, i) => (0, exports.getPriorityFeeEstimate)(provider.connection, tx.serialize(), i == currI ? priorityLevel : origPriorityLevel)));
            for (let i = 0; i < feeEstimates.length; i += 1) {
                const feeEstimate = feeEstimates[i];
                const computePriceIx = web3_js_1.ComputeBudgetProgram.setComputeUnitPrice({
                    // @ts-ignore
                    microLamports: Math.trunc(feeEstimate.priorityFeeEstimate),
                });
                const tx = txnsSigned[i];
                tx.add(computePriceIx);
                // resign
                const signers = ixAndSigners[i].signers;
                if (signers.length > 0) {
                    tx.partialSign(...signers);
                }
            }
        }
        log.debug('Sending multiple transactions...');
        try {
            return yield promiseAllInOrder(txnsSigned.map((txn, i) => () => __awaiter(void 0, void 0, void 0, function* () {
                try {
                    const { txid } = yield sendAndConfirmWithRetry(provider.connection, txn.serialize(), {
                        skipPreflight: true,
                    }, finality, timeout);
                    txSigs[currI] = txid;
                    currI += 1;
                    return txid;
                }
                catch (e) {
                    log.error(`txidx ${currI} err ${e}`);
                    throw e;
                }
            })));
        }
        catch (e) {
            log.error(e);
            const wrappedE = anchorError_1.ProgramError.parse(e, idlErrors);
            throw wrappedE == null ? e : wrappedE;
        }
    });
    const recur = (instructionGroups, signerGroups, priorityLevel) => __awaiter(void 0, void 0, void 0, function* () {
        try {
            yield sendMultipleTxs(instructionGroups, signerGroups, priorityLevel);
            return;
        }
        catch (e) {
            if (e instanceof anchorError_1.ProgramError) {
                log.error(`aborting. got program error ${e}`);
                throw e;
            }
            // @ts-ignore
            let baseMsg = `failed at prio ${priorityLevel}: ${e} ${e.stack}`;
            let newPrio;
            if (priorityLevel === 'NONE') {
                newPrio = 'Low';
            }
            else if (priorityLevel == 'Low') {
                newPrio = 'Medium';
            }
            else if (priorityLevel == 'Medium') {
                newPrio = 'High';
            }
            else if (priorityLevel == 'High') {
                newPrio = 'VeryHigh';
            }
            const msg = `${baseMsg}. ${newPrio ? `retrying at ${newPrio}` : `hard failed`}`;
            log.info(msg);
            if (!newPrio) {
                throw e;
            }
            yield recur(instructionGroups.slice(currI), signerGroups.slice(currI), newPrio);
            return;
        }
    });
    yield recur(instructionGroups, signerGroups, priorityLevel);
    return txSigs;
});
exports.sendMultipleInstructionsWithPrioFee = sendMultipleInstructionsWithPrioFee;
const SEND_TRANSACTION_INTERVAL = 10;
const awaitTransactionSignatureConfirmation = (txid, timeoutMillis, connection, commitment = 'recent', queryStatus = false, log = logger_1.consoleLogger) => __awaiter(void 0, void 0, void 0, function* () {
    let done = false;
    let status = {
        slot: 0,
        confirmations: 0,
        err: null,
    };
    let subId = 0;
    status = yield new Promise((resolve, reject) => __awaiter(void 0, void 0, void 0, function* () {
        let t;
        function setDone() {
            done = true;
            clearTimeout(t);
        }
        t = setTimeout(() => {
            if (done) {
                return;
            }
            setDone();
            log.error('Rejecting for timeout...');
            reject({ timeout: true });
        }, timeoutMillis);
        try {
            subId = connection.onSignature(txid, (result, context) => {
                status = {
                    err: result.err,
                    slot: context.slot,
                    confirmations: 0,
                };
                setDone();
                if (result.err) {
                    log.error('Rejected via websocket', result.err);
                    reject(status);
                }
                else {
                    resolve(status);
                }
            }, commitment);
        }
        catch (e) {
            log.error('WS error in setup', txid, e);
            if (!queryStatus) {
                reject(e);
            }
        }
        while (!done && queryStatus) {
            ;
            (() => __awaiter(void 0, void 0, void 0, function* () {
                try {
                    const signatureStatuses = yield connection.getSignatureStatuses([
                        txid,
                    ]);
                    status = signatureStatuses && signatureStatuses.value[0];
                    if (!done) {
                        if (!status) {
                        }
                        else if (status.err) {
                            log.error('REST error for', txid, status);
                            setDone();
                            reject(status.err);
                        }
                        else if (!status.confirmations && !status.confirmationStatus) {
                            log.warn('REST no confirmations for', txid, status);
                        }
                        else {
                            log.info('REST confirmation for', txid, status);
                            if (!status.confirmationStatus ||
                                status.confirmationStatus == commitment) {
                                setDone();
                                resolve(status);
                            }
                        }
                    }
                }
                catch (e) {
                    if (!done) {
                        log.error('REST connection error: txid', txid, e);
                    }
                }
            }))();
            yield sleep(2000);
        }
    }));
    if (
    //@ts-ignore
    connection._signatureSubscriptions &&
        //@ts-ignore
        connection._signatureSubscriptions[subId]) {
        connection.removeSignatureListener(subId);
    }
    done = true;
    return status;
});
exports.awaitTransactionSignatureConfirmation = awaitTransactionSignatureConfirmation;
function simulateTransaction(connection, transaction, commitment) {
    return __awaiter(this, void 0, void 0, function* () {
        // @ts-ignore
        transaction.recentBlockhash = yield connection._recentBlockhash(
        // @ts-ignore
        connection._disableBlockhashCaching);
        const signData = transaction.serializeMessage();
        // @ts-ignore
        const wireTransaction = transaction._serialize(signData);
        const encodedTransaction = wireTransaction.toString('base64');
        const config = { encoding: 'base64', commitment };
        const args = [encodedTransaction, config];
        // @ts-ignore
        const res = yield connection._rpcRequest('simulateTransaction', args);
        if (res.error) {
            throw new Error('failed to simulate transaction: ' + res.error.message);
        }
        return res.result;
    });
}
// DO NOT CHANGE THE NUMBERS BELOW:
// if you have retry logic and its set to retry after some N sec wait time << the lowest range of confirmation times
// (400-600ms/slot expiring after 151 slots -> 400 * 151 = 60.4s. so lets say you set it to 30s)
// you could wind up double mining your tx because you assumed the previous retry iteration timed out when in reality
// the tx was actually mined
const HIGH_END_TIMEOUT_MILLIS = 91 * 1000; // 91s -> 600ms high end slot time * 151 slots = 90.4s
const MAX_TIMEOUT_MILLIS = 120 * 1000; // 120s -> not sure where this num comes from
/*
    A validator has up to 120s to accept the transaction and send it into a block.
    If it doesn’t happen within that timeframe, your transaction is dropped and you’ll need
    to send the transaction again. You can get the transaction signature and periodically
    Ping the network for that transaction signature. If you never get anything back,
    that means it’s definitely been dropped. If you do get a response back, you can keep pinging
    that means it’s definitely been dropped. If you do get a response back, you can keep pinging
    until it’s gone to a confirmed status to move on.
  */
function sendAndConfirmWithRetry(connection, txn, sendOptions, commitment, timeoutMillis = HIGH_END_TIMEOUT_MILLIS, log = logger_1.consoleLogger, maxSupportedTransactionVersion) {
    var _a, _b;
    return __awaiter(this, void 0, void 0, function* () {
        let done = false;
        let slot = 0;
        const txid = yield connection.sendRawTransaction(txn, sendOptions);
        const startTime = Date.now();
        (() => __awaiter(this, void 0, void 0, function* () {
            while (!done && Date.now() - startTime < timeoutMillis) {
                log.debug(`sending raw tx now - startTime ${Date.now() - startTime} < ${timeoutMillis}`);
                yield connection.sendRawTransaction(txn, sendOptions);
                yield sleep(500);
            }
        }))();
        try {
            const confirmation = yield (0, exports.awaitTransactionSignatureConfirmation)(txid, timeoutMillis, connection, commitment, true);
            if (!confirmation)
                throw new Error('Timed out awaiting confirmation on transaction');
            if (confirmation.err) {
                const tx = yield connection.getTransaction(txid, {
                    commitment: 'confirmed',
                    maxSupportedTransactionVersion,
                });
                log.error((_b = (_a = tx === null || tx === void 0 ? void 0 : tx.meta) === null || _a === void 0 ? void 0 : _a.logMessages) === null || _b === void 0 ? void 0 : _b.join('\n'));
                log.error(confirmation.err);
                throw new Error('Transaction failed: Custom instruction error');
            }
            slot = (confirmation === null || confirmation === void 0 ? void 0 : confirmation.slot) || 0;
        }
        catch (err) {
            log.error('Error caught', err);
            if (err.timeout) {
                throw new Error('Timed out awaiting confirmation on transaction');
            }
            const tx = yield connection.getTransaction(txid, {
                commitment: 'confirmed',
                maxSupportedTransactionVersion,
            });
            if (tx && tx.meta && tx.meta.logMessages) {
                log.error(tx.meta.logMessages.join('\n'));
            }
            if (err.err) {
                throw err.err;
            }
            throw err;
        }
        finally {
            done = true;
        }
        log.info('Latency', txid, Date.now() - startTime);
        return { txid };
    });
}
exports.sendAndConfirmWithRetry = sendAndConfirmWithRetry;
function stringToTransaction(solanaTransaction) {
    return web3_js_1.Transaction.from(Buffer.from(solanaTransaction));
}
exports.stringToTransaction = stringToTransaction;
function bufferToTransaction(solanaTransaction) {
    return web3_js_1.Transaction.from(solanaTransaction);
}
exports.bufferToTransaction = bufferToTransaction;
function withRetries(tries, input, log = logger_1.consoleLogger) {
    return __awaiter(this, void 0, void 0, function* () {
        for (let i = 0; i < tries; i++) {
            try {
                return yield input();
            }
            catch (e) {
                log.debug(`Retrying ${i}...`, e);
            }
        }
        throw new Error('Failed after retries');
    });
}
const TX_BATCH_SIZE = 200;
function bulkSendTransactions(provider, txs, onProgress, triesRemaining = 10) {
    return __awaiter(this, void 0, void 0, function* () {
        let ret = [];
        // attempt to chunk by blockhash bounds (so signing doesn't take too long)
        for (let chunk of (0, accountFetchCache_1.chunks)(txs, TX_BATCH_SIZE)) {
            const thisRet = [];
            // Continually send in bulk while resetting blockhash until we send them all
            while (true) {
                const recentBlockhash = yield withRetries(5, () => provider.connection.getLatestBlockhash('confirmed'));
                const signedTxs = yield Promise.all(chunk.map((tx) => __awaiter(this, void 0, void 0, function* () {
                    tx.recentBlockhash = recentBlockhash.blockhash;
                    // @ts-ignore
                    const signed = yield provider.wallet.signTransaction(tx);
                    tx.signatures[0].signature;
                    return signed;
                })));
                const txsWithSigs = signedTxs.map((tx, index) => {
                    return {
                        transaction: chunk[index],
                        sig: bs58_1.default.encode(tx.signatures[0].signature),
                    };
                });
                const confirmedTxs = yield bulkSendRawTransactions(provider.connection, signedTxs.map((s) => s.serialize()), (_a) => {
                    var { totalProgress } = _a, rest = __rest(_a, ["totalProgress"]);
                    return onProgress &&
                        onProgress(Object.assign(Object.assign({}, rest), { totalProgress: totalProgress + ret.length + thisRet.length }));
                }, recentBlockhash.lastValidBlockHeight, 
                // Hail mary, try with preflight enabled. Sometimes this causes
                // errors that wouldn't otherwise happen
                triesRemaining != 1);
                thisRet.push(...confirmedTxs);
                if (confirmedTxs.length == signedTxs.length) {
                    break;
                }
                const retSet = new Set(thisRet);
                chunk = txsWithSigs
                    .filter(({ sig }) => !retSet.has(sig))
                    .map(({ transaction }) => transaction);
                triesRemaining--;
                if (triesRemaining <= 0) {
                    throw new Error(`Failed to submit all txs after blockhashes expired, ${signedTxs.length - confirmedTxs.length} remain`);
                }
            }
            ret.push(...thisRet);
        }
        return ret;
    });
}
exports.bulkSendTransactions = bulkSendTransactions;
// Returns the list of succesfully sent txns
// NOTE: The return signatures are ordered by confirmation, not by order they are passed
// This list should be in order. Seom txns may fail
// due to blockhash exp
function bulkSendRawTransactions(connection, txs, onProgress, lastValidBlockHeight, skipPreflight = true, log = logger_1.consoleLogger) {
    return __awaiter(this, void 0, void 0, function* () {
        const txBatchSize = TX_BATCH_SIZE;
        let totalProgress = 0;
        const ret = [];
        if (!lastValidBlockHeight) {
            const blockhash = yield withRetries(5, () => connection.getLatestBlockhash('confirmed'));
            lastValidBlockHeight = blockhash.lastValidBlockHeight;
        }
        for (let chunk of (0, accountFetchCache_1.chunks)(txs, txBatchSize)) {
            let currentBatchProgress = 0;
            let pendingCount = chunk.length;
            let txids = [];
            let lastRetry = 0;
            while (pendingCount > 0) {
                if ((yield withRetries(5, () => connection.getBlockHeight())) >
                    lastValidBlockHeight) {
                    return ret;
                }
                // only resend txs every 4s
                if (lastRetry < new Date().valueOf() - 4 * 1000) {
                    lastRetry = new Date().valueOf();
                    txids = [];
                    for (const tx of chunk) {
                        const txid = yield connection.sendRawTransaction(tx, {
                            skipPreflight,
                        });
                        txids.push(txid);
                    }
                }
                const statuses = yield getAllTxns(connection, txids);
                const completed = statuses.filter((status) => status !== null);
                totalProgress += completed.length;
                currentBatchProgress += completed.length;
                onProgress &&
                    onProgress({
                        totalProgress: totalProgress,
                        currentBatchProgress: currentBatchProgress,
                        currentBatchSize: txBatchSize,
                    });
                const failures = completed
                    .map((status) => { var _a; return status !== null && ((_a = status.meta) === null || _a === void 0 ? void 0 : _a.err); })
                    .filter(truthy);
                if (failures.length > 0) {
                    log.error(failures);
                    throw new Error('Failed to run txs');
                }
                ret.push(...txids
                    .map((txid, idx) => (statuses[idx] == null ? null : txid))
                    .filter(truthy));
                chunk = chunk.filter((_, index) => statuses[index] === null);
                txids = txids.filter((_, index) => statuses[index] === null);
                pendingCount -= completed.length;
                yield sleep(1000); // Wait one seconds before querying again
            }
        }
        return ret;
    });
}
exports.bulkSendRawTransactions = bulkSendRawTransactions;
const MAX_GET_SIGNATURE_STATUSES_QUERY_ITEMS = 200;
function getAllTxns(connection, txids) {
    return __awaiter(this, void 0, void 0, function* () {
        return (yield Promise.all((0, accountFetchCache_1.chunks)(txids, MAX_GET_SIGNATURE_STATUSES_QUERY_ITEMS).map((txids) => connection.getTransactions(txids, {
            maxSupportedTransactionVersion: 0,
            commitment: 'confirmed',
        })))).flat();
    });
}
class SendIxError extends Error {
    constructor(originalError) {
        var _a;
        //The newlines don't actually show up correctly in chai's assertion error, but at least
        // we have all the information and can just replace '\n' with a newline manually to see
        // what's happening without having to change the code.
        const logs = ((_a = originalError.logs) === null || _a === void 0 ? void 0 : _a.join('\n')) || 'error had no logs';
        super(originalError.message + '\nlogs:\n' + logs);
        this.stack = originalError.stack;
        this.logs = logs;
    }
}
//# sourceMappingURL=transaction.js.map